/*******************************************************************************
 * Copyright (C) 2018 THL A29 Limited, a Tencent company.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Tencent - support multiple contexts
 ******************************************************************************/
package com.eclipsesource.v8;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;

/**
 * Created by LiYongjie on 2018/4/16.
 */
public final class MultiContextNodeJS {
    private static final String TMP_JS_EXT          = ".js.tmp";
    private static final String GLOBAL              = "global";
    private static final String STARTUP_CALLBACK    = "__run";
    private static final String STARTUP_SCRIPT      = "global." + STARTUP_CALLBACK + "(require, exports, module, __filename, __dirname);";
    private static final String STARTUP_SCRIPT_NAME = "startup";

    private final MultiContextV8 mv8;
    private final V8Context mainContext;
    private V8Function require;

    public static MultiContextNodeJS createMultiContextNodeJS(int contextTag) {
        MultiContextV8 mv8 = MultiContextV8.createMultiContextV8(GLOBAL);

        /**
         * node will be started in this context
         */
        V8Context mainContext = mv8.createContext(contextTag);
        final MultiContextNodeJS node = new MultiContextNodeJS(mv8, mainContext);

        mainContext.registerJavaMethod(new JavaVoidCallback() {

            @Override
            public void invoke(final V8Object receiver, final V8Array parameters) {
                V8Function require = (V8Function) parameters.get(0);
                try {
                    node.init(require.twin());
                } finally {
                    require.release();
                }
            }
        }, STARTUP_CALLBACK);

        try {
            File startupScript = createTemporaryScriptFile(STARTUP_SCRIPT, STARTUP_SCRIPT_NAME);
            try {
                mv8.getV8().createNodeRuntime(startupScript.getAbsolutePath());
            } finally {
                startupScript.delete();
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return node;
    }

    public V8Context getMainContext() {
        return mainContext;
    }

    public MultiContextV8 getRuntime() {
        return mv8;
    }

    /**
     * Handles the next message in the message loop. Returns True
     * if there are more messages to handle, false otherwise.
     *
     * @return True if there are more messages to handle, false otherwise.
     */
    public boolean handleMessage() {
        mv8.getV8().checkThread();
        return mv8.getV8().pumpMessageLoop();
    }

    /**
     * Returns true if there are more messages to process, false otherwise.
     *
     * @return True if there are more messages to process, false otherwise.
     */
    public boolean isRunning() {
        mv8.getV8().checkThread();
        return mv8.getV8().isRunning();
    }

    public void release() {
        mv8.getV8().checkThread();

        if (!require.isReleased()) {
            require.release();
        }

        mainContext.release();
        mv8.release();
    }

    private MultiContextNodeJS(MultiContextV8 mv8, V8Context mainContext) {
        this.mv8 = mv8;
        this.mainContext = mainContext;
    }

    private void init(final V8Function require) {
        this.require = require;
    }

    private static File createTemporaryScriptFile(final String script, final String name) throws IOException {
        File tempFile = File.createTempFile(name, TMP_JS_EXT);
        PrintWriter writer = new PrintWriter(tempFile, "UTF-8");
        try {
            writer.print(script);
        } finally {
            writer.close();
        }
        return tempFile;
    }
}
