/*******************************************************************************
 * Copyright (C) 2018 THL A29 Limited, a Tencent company.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Tencent - support multiple contexts
 ******************************************************************************/
package com.eclipsesource.v8;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.nio.ByteBuffer;

/**
 * Created by LiYongjie on 2018/3/9.
 */
/**package**/class V8ContextWrapper implements InvocationHandler {
    private static final String TAG = "V8ContextWrapper";

    private final long v8ContextPtr;
    private final MultiContextV8 multiContextV8;
    private final V8ContextImpl v8ContextImpl;
    private final V8Context proxy;

    V8ContextWrapper(final MultiContextV8 multiContextV8, final long v8ContextPrt) {
        this.multiContextV8 = multiContextV8;
        this.v8ContextPtr = v8ContextPrt;

        enterContext();

        this.v8ContextImpl = new V8ContextImpl(v8ContextPrt);
        this.proxy = (V8Context) Proxy.newProxyInstance(V8ContextWrapper.class.getClassLoader(), new Class[]{V8Context.class}, this);
    }

    public V8Context context() {
        return proxy;
    }

    protected long getPtr() {
        return v8ContextPtr;
    }

    private void enterContext() {
        multiContextV8.enterContext(this);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        V8ContextWrapper that = (V8ContextWrapper) o;

        return v8ContextPtr == that.v8ContextPtr;
    }

    @Override
    public int hashCode() {
        return (int) (v8ContextPtr ^ (v8ContextPtr >>> 32));
    }

    private final class V8ContextImpl extends V8Object implements V8Context {
        private final long ptr;
        V8ContextImpl(long ptr) {
            super(multiContextV8.getV8());
            objectHandle = v8.getGlobalObject();
            this.ptr = ptr;
        }

        @Override
        public void executeVoidScript(String script) {
            v8.executeVoidScript(script);
        }

        @Override
        public void executeVoidScript(String script, String scriptName, int lineNumber) {
            v8.executeVoidScript(script, scriptName, lineNumber);
        }

        @Override
        public int executeIntegerScript(String script) {
            return v8.executeIntegerScript(script);
        }

        @Override
        public int executeIntegerScript(String script, String scriptName, int lineNumber) {
            return v8.executeIntegerScript(script, scriptName, lineNumber);
        }

        @Override
        public double executeDoubleScript(String script) {
            return v8.executeDoubleScript(script);
        }

        @Override
        public double executeDoubleScript(String script, String scriptName, int lineNumber) {
            return v8.executeDoubleScript(script, scriptName, lineNumber);
        }

        @Override
        public String executeStringScript(String script) {
            return v8.executeStringScript(script);
        }

        @Override
        public String executeStringScript(String script, String scriptName, int lineNumber) {
            return v8.executeStringScript(script, scriptName, lineNumber);
        }

        @Override
        public boolean executeBooleanScript(String script) {
            return v8.executeBooleanScript(script);
        }

        @Override
        public boolean executeBooleanScript(String script, String scriptName, int lineNumber) {
            return v8.executeBooleanScript(script, scriptName, lineNumber);
        }

        @Override
        public V8Array executeArrayScript(String script) {
            return v8.executeArrayScript(script);
        }

        @Override
        public V8Array executeArrayScript(String script, String scriptName, int lineNumber) {
            return v8.executeArrayScript(script, scriptName, lineNumber);
        }

        @Override
        public Object executeScript(String script) {
            return v8.executeScript(script);
        }

        @Override
        public Object executeScript(String script, String scriptName, int lineNumber) {
            return v8.executeScript(script, scriptName, lineNumber);
        }

        @Override
        public V8Object executeObjectScript(String script) {
            return v8.executeObjectScript(script);
        }

        @Override
        public V8Object executeObjectScript(String script, String scriptName, int lineNumber) {
            return v8.executeObjectScript(script, scriptName, lineNumber);
        }

        @Override
        public void shareObject(String jsObjectName, V8Context targetContext) {
            shareObjectImpl(jsObjectName, targetContext, null);
        }

        @Override
        public void shareObject(String jsObjectName, V8Context targetContext, String targetJsObjectRename) {
            shareObjectImpl(jsObjectName, targetContext, targetJsObjectRename);
        }

        @Override
        public V8Object newV8Object() {
            return new V8Object(v8);
        }

        @Override
        public V8Array newV8Array() {
            return new V8Array(v8);
        }

        @Override
        public V8ArrayBuffer newV8ArrayBuffer(int capacity) {
            return new V8ArrayBuffer(v8, capacity);
        }

        @Override
        public V8ArrayBuffer newV8ArrayBuffer(ByteBuffer byteBuffer) {
            return new V8ArrayBuffer(v8, byteBuffer);
        }

        @Override
        public V8Function newV8Function(JavaCallback javaCallback) {
            return new V8Function(v8, javaCallback);
        }

        @Override
        public V8TypedArray newV8TypedArray(V8ArrayBuffer buffer, int type, int offset, int size) {
            return new V8TypedArray(v8, buffer, type, offset, size);
        }

        @Override
        public V8Object getGlobalObject() {
            return this;
        }

        @Override
        public void release() {
            if (isReleased()) {
                return;
            }

            v8.releaseObjRef(this);

            multiContextV8.releaseContext(V8ContextWrapper.this);
            released = true;
        }

        @Override
        public long getPtr() {
            return ptr;
        }

        private void shareObjectImpl(String jsObjectName, V8Context targetContext, String targetJsObjectRename) {
            checkReleased();
            v8.checkThread();
            if(jsObjectName == null || jsObjectName.isEmpty()) {
                return;
            }

            if (targetContext == null) {
                return;
            }

            V8Object source = getObject(jsObjectName);

            if (targetJsObjectRename == null || targetJsObjectRename.isEmpty()) {
                targetJsObjectRename = jsObjectName;
            }
            targetContext.add(targetJsObjectRename, source);

            source.release();
        }
    }

    @Override
    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        v8ContextImpl.checkReleased();
        enterContext();
        try {
            return method.invoke(v8ContextImpl, args);
        } catch (InvocationTargetException e) {
            throw e.getCause();
        }
    }
}
